# pylint: disable=invalid-name

'''
functions for manipulating expression profile matrices
'''

import rpy2.robjects as ro

class Expression(object):
    '''
    store and manipulate expression profile matrix
    '''
    def __init__(self, data):
        self.profiles = data
        self.size = len(self.profiles)
        return None


    def genes(self):
        '''
        return genes (row names)
        '''
        return self.profiles.rownames


    def nrow(self):
        '''
        return number of rows
        '''
        return self.profiles.nrow


    def ncol(self):
        '''
        return number of columns
        '''
        return self.profiles.ncol


    def samples(self):
        '''
        return column names (samples)
        '''
        return self.profiles.colnames


    def expression(self):
        '''
        wrapper for accessing self.profiles
        '''
        return self.profiles


    def gene_expression(self, genes):
        '''
        subsets expression data
        returning expression for list of genes
        '''
        return self.profiles.rx(ro.StrVector(genes), True)


    def residual_expression(self, unclassifiedGenes):
        '''
        subsets expression data
        returning expression for only residuals to the fit
        '''
        if unclassifiedGenes is None:
            return None
        else:
            return self.gene_expression(unclassifiedGenes)


    def fit_expression(self, fitGenes):
        '''
        subsets expression data
        return only genes that passed the
        goodness of fit assessment
        '''
        if fitGenes is None:
            return None
        else:
            return self.gene_expression(fitGenes)



# pylint: disable=invalid-name
# pylint: disable=no-self-use
'''
wgcna functions
'''

import logging
from collections import OrderedDict
import rpy2.robjects as ro
from .r.imports import base, wgcna, rsnippets, stats
from .r.manager import RManager

class WgcnaManager(RManager):
    '''
    wrappers for running WGCNA functions
    an extension of the RManager
    '''
    def __init__(self, data, params, debug=False):
        RManager.__init__(self, data, params)
        self.logger = logging.getLogger('iterativeWGCNA.WgcnaManager')
        self.adjacencyMatrix = None
        self.TOM = None
        self.dissimilarityMatrix = None
        self.geneTree = None
        self.moduleColors = None
        self.debug = debug
        return None


    def set_module_colors(self, modules):
        '''
        set module colors from dict of module properties
        expects dict to have a 'color' sub-dict
        '''
        self.moduleColors = OrderedDict((module, values['color']) \
                                            for module, values in modules.items())


    def blockwise_modules(self):
        '''
        run blockwise WGCNA
        '''
        self.params['datExpr'] = self.transpose_data()
        blocks = wgcna().blockwiseModules(**self.params)
        self.collect_garbage()
        return blocks


    def collect_garbage(self):
        '''
        run WGCNA garbage collection
        '''
        wgcna().collectGarbage()


    def adjacency(self, networkType='signed', removeNegatives=False, removeSelfReferences=False):
        '''
        calculate adjacency matrix; from pearson correlation
        '''

        adjParams = {}
        adjParams['power'] = self.params['power'] if 'power' in self.params else 6
        adjParams['corFnc'] = 'cor'
        adjParams['corOptions'] = "use='p'"
        adjParams['type'] = networkType
        adjParams['datExpr'] = self.transpose_data()

        self.adjacencyMatrix = wgcna().adjacency(**adjParams)
        self.collect_garbage()
        if removeNegatives:
            self.adjacencyMatrix = rsnippets.filterByThreshold(self.adjacencyMatrix, 0)

        if removeSelfReferences:
            self.adjacencyMatrix = rsnippets.diag(self.adjacencyMatrix, 0)

        self.collect_garbage()


    def TOM_dist(self):
        '''
        calculate Topological Overlap Matrix from adjacency matrix
        '''
        self.TOM = wgcna().TOMdist(self.adjacencyMatrix)
        self.collect_garbage()


    def TOM_similarity_from_expr(self):
        '''
        calculate Topological Overlap Matrix from expression data
        '''
        funcParams = {}
        funcParams['power'] = self.params['power'] if 'power' in self.params else 6
        funcParams['datExpr'] = self.transpose_data()
        self.TOM = wgcna().TOMsimilarityFromExpr(**funcParams)
        self.collect_garbage()


    def plot_network_heatmap(self, membership, title, useTOM=False):
        '''
        plot network heatmap
        recapitulates WGCNA plotNetworkOverview to work around
        cutree issues
        uses pheatmap
        '''

        # TODO fix useTOM option for drawing with pheatmap and use of similarity
        # instead of dissimilarity

        # if useTOM:
            # self.TOM_similarity_from_expr()
            # self.dissimilarityMatrix = rsnippets.dissMatrix(self.TOM)
			# raising TOM to power of 7 recommended by WGCNA documentation
            # self.dissimilarityMatrix = rsnippets.powerWeightMatrix(self.TOM, 7)
        #else:

        self.adjacency()

        annotation = self.heatmap_annotation_data_frame(['Module'], membership)
        annotationKey = self.heatmap_annotation_key('Module', self.moduleColors)
        # manager = RManager(self.dissimilarityMatrix, None)
        self.heatmap(clusterCols=True, params={'scale': 'none',
                                               'mat': self.adjacencyMatrix,
                                               'show_colnames': False,
                                               'color': rsnippets.WhYlRed(),
                                               'annotation_col': annotation,
                                               'annotation_row': annotation,
                                               'annotation_colors': annotationKey})


    def generate_gene_tree(self):
        '''
        generate hierarchical cluster from dissimilarity matrix
        '''
        distMatrix = stats().as_dist(self.dissimilarityMatrix)
        self.geneTree = stats().hclust(distMatrix, method="average")


    def plot_network_overview(self, moduleColors, title, useTOM=False):
        '''
        wrapper for TOMplot which provides a graphical representation
        of the Topological Overlap Matrix or correlation matrix using a heatmap
        and hierarchical clustering dendrogram annotated by
        module colors
        '''

        if useTOM:
            self.TOM_similarity_from_expr()
            self.dissimilarityMatrix = rsnippets.dissMatrix(self.TOM)
            self.dissimilarityMatrix = rsnippets.powerWeightMatrix(self.TOM, 7)
        else:
            self.adjacency()
            self.dissimilarityMatrix = rsnippets.dissMatrix(self.adjacencyMatrix)

        # self.dissimilarityMatrix = rsnippets.diag(self.dissimilarityMatrix, ro.NA_Integer)
        self.generate_gene_tree()

        params = {}
        params['dissim'] = self.dissimilarityMatrix
        params['dendro'] = self.geneTree
        params['Colors'] = moduleColors
        params['main'] = title
        wgcna().TOMplot(**params)


    def module_eigengenes(self, membership):
        '''
        wrapper for moduleEigengenes function
        calculates eigengenes from profiles &
        module membership (gene -> membership dict)
        '''

        if self.debug:
            self.logger.debug("Running WGCNA moduleEigengenes function")
            self.logger.debug("Module assignments:")
            self.logger.debug(membership)

        params = {}
        params['softPower'] = self.params['power'] if 'power' in self.params else 6
        params['expr'] = base().as_data_frame(self.transpose_data())

        if self.debug:
            self.logger.debug("Converting membership list to ro.StrVector; see R-log")
            ro.r("print('Converting membership list to ro.StrVector for WGCNA moduleEigengenes:')")

        params['colors'] = ro.StrVector(list(membership))

        if self.debug:
            self.logger.debug(params['colors'])

        return wgcna().moduleEigengenes(**params)


    def plot_eigengene_network(self):
        '''
        wrapper for plotEigengeneNetworks function
        plots an eigengene network
        '''
        params = {}
        params['multiME'] = base().as_data_frame(self.transpose_data())
        params['setLabels'] = ''
        params['marDendro'] = ro.IntVector([0, 4, 1, 2])
        params['marHeatmap'] = ro.IntVector([3, 4, 1, 2])
        params['cex.lab'] = 0.8
        params['xLabelsAngle'] = 90
        params['colorLabels'] = False
        params['signed'] = True

        wgcna().plotEigengeneNetworks(**params)




# pylint: disable=invalid-name
# pylint: disable=unused-import
'''
manage eigengenes
'''

from __future__ import print_function

import logging

import rpy2.robjects as ro
from .r.imports import base, stats, rsnippets
from .io.utils import write_data_frame
from .wgcna import WgcnaManager

class Eigengenes(object):
    '''
    manage and manipulate eigengene matrices
    '''

    def __init__(self, matrix=None, debug=False):
        self.debug = debug
        self.logger = logging.getLogger('iterativeWGCNA.Eigengenes')
        self.matrix = matrix


    def extract_from_blocks(self, iteration, blocks, samples):
        '''
        extract eigenenges from blockwise WGCNA results
        '''
        self.matrix = rsnippets.extractEigengenes(iteration, blocks, samples)


    def samples(self):
        '''
        return sample names
        '''
        return self.matrix.names


    def nrows(self):
        '''
        wrapper for returning number of rows in
        the eigengene matrix
        '''

        return self.matrix.nrow


    def load_matrix_from_file(self, fileName):
        '''
        loads eigengenes from file into an R DataFrame
        '''
        self.matrix = ro.DataFrame.from_csvfile(fileName, sep='\t',
                                                header=True, row_names=1)


    def write(self, prefix=''):
        '''
        writes the eigengene matrix to file
        '''
        fileName = prefix + 'eigengenes.txt'
        write_data_frame(self.matrix, fileName, 'Module')


    def similarity(self, module=None):
        '''
        calculate similarity between eigengene for a specific
        module and all the other eigengenes

        if no module is specified, calculate the similarity matrix
        between all eigengenes
        '''
        if module is None:
            sim = base().as_data_frame(stats().cor(base().t(self.matrix)))
        else:
            sim = base().as_data_frame(stats().cor(base().t(self.matrix), \
                                            base().t(self.matrix.rx(module, True))))
        return sim


    def correlation(self, m1, m2):
        '''
        calculate correlation between two module eigengenes
        '''
        e1 = self.get_module_eigengene(m1)
        e2 = self.get_module_eigengene(m2)
        cor = base().as_data_frame(stats().cor(base().t(e1), base().t(e2)))
        cor = round(cor.rx(1, 1)[0], 1)
        return cor


    def equal(self, m1, m2, threshold=0.0):
        '''
        check if 2 module eigengenes are "equivalent"
        (1 - correlation <= threshold)
        '''
        cor = self.correlation(m1, m2)
        return 1.0 - cor <= threshold


    def get_module_eigengene(self, module):
        '''
        return a module eigengene
        '''
        return self.matrix.rx(module, True)


    def extract_subset(self, modules):
        '''
        return a submatrix
        '''
        if self.debug:
            self.logger.debug("Extracting eigengenes for the following modules:")
            self.logger.debug(modules)

        if self.debug:
            self.logger.debug("Converting module list to ro.StrVector; see R-log")
            ro.r("print('Converting module list to ro.StrVector to extract eigengenes:')")

        vector = ro.StrVector(modules)

        if self.debug:
            self.logger.debug(vector)

        if self.debug:
            self.logger.debug("Extracted submatrix, see R-log")
            ro.r("print('Extracted eigengene submatrix:')")


        newMatrix = self.matrix.rx(vector, True)

        if self.debug:
            self.logger.debug(newMatrix)

        return newMatrix



    def is_empty(self):
        '''
        return True if matrix is empty
        '''
        return self.matrix.nrow == 0


    def update_to_subset(self, modules):
        '''
        update matrix to subset specified by modules
        '''
        self.matrix = self.extract_subset(modules)


    def recalculate(self, profiles, membership, power=6):
        '''
        recalculate eigengenes given membership
        and profiles
        '''
        manager = WgcnaManager(profiles, {'power':power}, debug=self.debug)

        self.matrix = rsnippets.extractRecalculatedEigengenes(
            manager.module_eigengenes(membership.values()),
            self.samples())



#!/usr/bin/env python2.7
"""
Perform iterative WGCNA analysis

python dependencies:
  * rpy2
  * matplotlib

R dependencies:
  * WGCNA
"""

from .cmlargs import parse_command_line_args
from .iterativeWGCNA import IterativeWGCNA

if __name__ == '__main__':
    args = parse_command_line_args()
    alg = IterativeWGCNA(args)
    alg.run()

__author__ = 'Emily Greenfest-Allen'
__copyright__ = 'Copyright 2016, University of Pennsylvania'




'''Rerun final module merge'''

# Installation workaround - see README
# import readline
#pylint: disable=invalid-name

from iterativeWGCNA.iterativeWGCNA import IterativeWGCNA
from iterativeWGCNA.cmlargs import parse_command_line_args

if __name__ == '__main__':
    args = parse_command_line_args(program='iterativeWGCNA: Adjust Merge',
                                   description='recompute final module merge from existing output')
    alg = IterativeWGCNA(args, report="merge")
    alg.merge_close_modules_from_output()

__author__ = 'Emily Greenfest-Allen'
__copyright__ = 'Copyright 2018, University of Pennsylvania'





# pylint: disable=invalid-name
# pylint: disable=no-self-use
'''
Wrappers for R functions;
performs conversions to rpy2.robjects
where required
'''

import logging
import rpy2.robjects as ro
from collections import OrderedDict

# import rpy2.rlike.container as rlc
from .imports import base, pheatmap, graphics, rsnippets

class RManager(object):
    '''
    wrappers for running R functions
    '''
    def __init__(self, data, params=None):
        self.logger = logging.getLogger('iterativeWGCNA.RManager')
        self.data = data
        if params is None:
            self.params = {}
        else:
            self.params = params

        return None


    def update_parameters(self, params):
        '''
        update/replace all parameters
        '''
        self.params = params


    def set_parameter(self, name, value):
        '''
        add or update a single parameter
        '''
        self.params[name] = value


    def remove_parameter(self, name):
        '''
        remove named parameter
        '''
        del self.params[name]


    def transpose_data(self):
        '''
        transpose the data frame (required for some WGCNA functions)
        '''
        return base().t(self.data)


    def log2(self):
        '''
        log2 data
        '''
        return base().log2(rsnippets.add(self.data, 1))


    def row_names(self):
        '''
        wrapper for getting
        row names (usually genes)
        '''
        return self.data.rownames


    def col_names(self):
        '''
        wrapper for getting column names
        (usually samples)
        '''
        return self.data.names


    def heatmap_annotation_data_frame(self, categories, annotation):
        '''
        takes a dict of gene->value and creates a data frame
        data frame
        assume annotation is an ordered dict
        updates column names to names
        '''
        df = base().as_data_frame(base().t(ro.DataFrame(annotation)))
        df.colnames = ro.StrVector(categories)
      
        return df


    def heatmap_annotation_key(self, name, colors):
        '''
        generates data frame for color key for the annotation
        from a dict
        '''
        keyColors = ro.StrVector([c for c in colors.values()])
        keyColors.names = colors.keys()
        key = OrderedDict()
        key[name] = keyColors

        return ro.ListVector(key)


    def heatmap(self, clusterCols=False, params=None):
        '''
        plot a heatmap with options specified in params
        (see pheatmap documentation for all options)
        '''
        self.params['mat'] = base().as_matrix(self.log2())
        self.params['border'] = ro.NA_Logical
        self.params['cluster_cols'] = clusterCols
        if clusterCols:
            self.params['clustering_distance_cols'] = 'correlation'
        self.params['clustering_distance_rows'] = 'correlation'
        self.params['show_rownames'] = True if self.data.nrow <= 50 else False
        self.params['scale'] = 'row'
        self.params['color'] = rsnippets.BlWhRed()

        if params is not None:
            self.params.update(params)

        pheatmap().pheatmap(**self.params)



    def barchart(self, params=None):
        '''
        barchart
        '''

        self.params['height'] = base().as_numeric(self.data)
        if params is not None:
            self.params.update(params)

        graphics().barplot(**self.params)


    def histogram(self, vline=None, params=None):
        '''
        plot histogram with vline at x=vline
        '''
        self.params['x'] = ro.FloatVector(self.data)
        self.params['labels'] = False

        if params is not None:
            self.params.update(params)

        graphics().hist(**self.params)

        if vline is not None:
            lineParams = {'v': vline, 'col': 'red'}
            graphics().abline(**lineParams)






#pylint: disable=anomalous-backslash-in-string
#pylint: disable=invalid-name

'''
functions for defining, parsing, processing,
and validating command line arguments
'''

import re
import argparse
from os import getcwd
from .io.utils import warning

def parameter_list(strValue):
    '''
    for argument parsing;
    converts a comma separated list of 'param=value' pairs
    into a parm:value hash
    '''

    params = {}
    pairs = strValue.split(',')

    for p in pairs:
        name, value = p.split('=')

        # Test and cast for booleans
        if value.upper() in ['TRUE', 'T']:
            params[name] = True
        elif value.upper() in ['FALSE', 'F']:
            params[name] = False
        # Test and cast for integer
        elif value.isdigit():
            params[name] = int(value)
        # Test and cast for float
        elif re.match("^\d+?\.\d+?$", value):
            params[name] = float(value)
        else:
            params[name] = value

    return params


def restricted_float(x):
    '''
    for argument parsing; restricts float value from 0 to 1
    '''
    x = float(x)
    if x < 0.0 or x > 1.0:
        raise argparse.ArgumentTypeError("%r not in range [0.0, 1.0]"%(x,))
    return x


def summaryHelpEpilog():
    '''
    text for help epilog for
    summary
    '''
    # TODO Help for summarize script
    return "COMING SOON"

def helpEpilog():
    '''
    text for help epilog
    '''

    inputFileFormatHelp = '''
------------------------------------------------------
Input File Format
------------------------------------------------------
iterativeWGCNA expects a tab-delimited text file containing
gene expression data arranged such that there is one
row per gene and one column per sample.  The first column
should contain unique gene identifiers.  For example:

GENE    Sample1    Sample2    Sample3
Gata1    500    715    1000
Phtf2    60    1000    1600
'''
    wgcnaParametersHelp = '''
------------------------------------------------------
WGCNA Parameters
------------------------------------------------------
iterativeWGCNA can accept any parameter valid for the WGCNA
blockwiseModules function.

See http://www.inside-r.org/packages/cran/wgcna/docs/blockwiseModules

To specify these parameters use the --wgcnaParameters flag followed by
a comma separated list of parameter=value pairs.  For example:

--wgcnaParameters maxBlockSize=5000,corType=bicor,power=10

sets the maximum block size to 5000 genes,
the correlation type to the biweight correlation,
and the power-law scaling factor (beta) to 10

If parameters are not specified, iterativeWGCNA uses the default WGCNA settings,
except for the following:

minModuleSize=20
saveTOMs=FALSE
minKMEtoStay=0.8
minCoreKME=0.8
networkType=signed
numericLabels=TRUE

'''

    return inputFileFormatHelp + '\n\n' + wgcnaParametersHelp


def parse_command_line_args(program='iterativeWGCNA', description='perform iterativeWGCNA analysis'):
    '''
    parse command line args
    '''

    parser = argparse.ArgumentParser(prog=program,
                                     description=description,
                                     epilog=helpEpilog(),
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-i', '--inputFile',
                        metavar='<gene expression file>',
                        help="full path to input gene expression file; "
                        + "if full path is not provided,\n"
                        + "assumes the file is in the working directory\n;"
                        + "see below for input file format",
                        required=True)

    parser.add_argument('-o', '--workingDir',
                        help="R working directory; where output will be saved",
                        metavar='<output dir>', 
                        default=getcwd())

    parser.add_argument('-v', '--verbose',
                        help="print status messages",
                        action='store_true')

    parser.add_argument('--debug',
                        help="print debugging messages",
                        action='store_true')
    
    parser.add_argument('-p', '--wgcnaParameters',
                        metavar='<param list>',
                        help="comma separated list of parameters to be passed to WGCNA's "
                        + "blockwiseModules function\n"
                        + "e.g., power=6,randomSeed=1234875\n"
                        + "see WGCNA manual & more info below",
                        type=parameter_list)

    parser.add_argument('--enableWGCNAThreads',
                        help="enable WGCNA to use threading;\nsee WGCNA manual",
                        action='store_true')

    parser.add_argument('--skipSaveBlocks',
                        help="do not save WGCNA blockwise modules for each iteration;\n"
                        + "NOTE: without blocks summary graphics cannot be generated.\n"
                        + "Also will not saveTOMs.",
                        action='store_true')

    parser.add_argument('--gzipTOMs',
                        help="gzip TOM RData files\n",
                        action='store_true')

    parser.add_argument('-f', '--finalMergeCutHeight',
                        help="cut height for final merge (after iterations are assembled)",
                        default=0.05,
                        metavar='<cut height>',
                        type=restricted_float)

    args = parser.parse_args()
    args.wgcnaParameters = set_wgcna_parameter_defaults(args.wgcnaParameters, args.skipSaveBlocks)

    return args


def set_wgcna_parameter_defaults(params, skipSaveBlocks):
    '''
    set default values for WGCNA blockwiseModules
    numericLabels = TRUE
    minKMEtoStay = 0.8
    '''

    if params is None:
        params = {}

    params['numericLabels'] = True # override user choice
    
    if 'networkType' not in params:
        params['networkType'] = 'signed'
    if 'minKMEtoStay' not in params:
        params['minKMEtoStay'] = 0.8
    if 'minCoreKME' not in params:
        params['minCoreKME'] = params['minKMEtoStay']
    if 'minModuleSize' not in params:
        params['minModuleSize'] = 20
    if 'reassignThreshold' not in params:
        params['reassignThreshold'] = 0.05 # 0.000001 # 1e-6
    if 'power' not in params:
        params['power'] = 6
    if 'saveTOMs' not in params:
        params['saveTOMs'] = False
    if skipSaveBlocks: # if blocks are not saved; TOMs are not saved
        params['saveTOMs'] = False

    return params


def parse_summary_command_line_args():
    '''
    parse command line args for summary
    '''

    parser = argparse.ArgumentParser(prog='iterativeWGCNA network summary',
                                     description="generate graphical results from interative WGCNA analysis",
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-i', '--inputFile',
                        metavar='<gene expression file>',
                        help="full path to input gene expression file; "
                        + "if full path is not provided,\n"
                        + "assumes the file is in the working directory\n;"
                        + "see below for input file format",
                        required=True)

    parser.add_argument('-o', '--workingDir',
                        help="R working directory; where output will be saved",
                        metavar='<output dir>',
                        default=getcwd())

    parser.add_argument('-v', '--verbose',
                        help="print status messages",
                        action='store_true')

    parser.add_argument('-p', '--power',
                        metavar='<power law beta>',
                        help="power law beta for weighting the adjacency matrix",
                        default=6,
                        type=int)

    parser.add_argument('--signed',
                        help="generate signed adjacency matrix?",
                        action='store_true')

    parser.add_argument('--minKMEtoStay',
                        help="provide minKMEtoStay used for network generation",
                        default=0.80,
                        metavar='<minKMEtoStay>',
                        type=restricted_float)

    parser.add_argument('--enableWGCNAThreads',
                        help="enable WGCNA to use threading;\nsee WGCNA manual",
                        action='store_true')

    parser.add_argument('--generateNetworkSummary',
                        metavar='<view type>',
                        choices=['all', 'network', 'input'],
                        help="generate summary overview of the network (dendrogram & heatmap):\n"
                        + "network - network comprised only of classified genes\n"
                        + "input - all genes, with classified highlighted by module assignments\n"
                        + "all - both views\n"
                        + "NOTE: all adjacency matrix calculations are\n"
                        + "done in one block and may fail due to memory allocation\n"
                        + "issues for large gene-sets")

    parser.add_argument('-e', '--edgeWeight',
                        metavar='<min edge weight>',
                        default=0.5,
                        help="min edge weight for network summary; filters for\n"
                        + "connections supported by a correlation >= threshold",
                        type=restricted_float)


    return parser.parse_args()





'''
I/O Utils
'''
from __future__ import print_function
from __future__ import with_statement

from sys import stderr
from math import isnan
import os
import re
from subprocess import check_call
import gzip

import rpy2.robjects as ro
from ..r.imports import rsnippets

def bulk_gzip(directory, pattern):
    '''
    gzip all files in the directory that match the pattern
    '''
    for filename in os.listdir(directory):
        if pattern in filename:
            with open(os.path.join(directory, filename), 'rb') as plain_file:
                with gzip.open(os.path.join(directory, filename + '.gz'), 'wb') as zip_file:
                    zip_file.writelines(plain_file)

            os.remove(os.path.join(directory, filename)) # remove the old file


def xstr(value):
    '''
    handle nulls/nan in string conversion
    '''
    if value is None:
        return ''
    if value == 'NULL':
        return ''
    if isnan(value):
        return 'NA'

    return str(value)


def warning(*objs):
    '''
    wrapper for writing to stderr
    '''
    print(*objs, file=stderr)
    stderr.flush()


def create_dir(dirName):
    '''
    check if directory exists in the path, if not create
    '''
    try:
        os.stat(dirName)
    except OSError:
        os.mkdir(dirName)

    return dirName


def write_data_frame(df, fileName, rowLabel):
    '''
    write data frame to file; creates new file
    if none exists, otherwise appends new eigengenes
    to existing file
    '''
    try:
        os.stat(fileName)
    except OSError:
        header = (rowLabel,) + tuple(df.colnames)
        with open(fileName, 'w') as f:
            print('\t'.join(header), file=f)
    finally:
        df.to_csvfile(fileName, quote=False, sep='\t', col_names=False, append=True)

        
def read_data(fileName):
    '''
    read gene expression data into a data frame
    and convert numeric (integer) data to real
    '''
    data = ro.DataFrame.from_csvfile(fileName, sep='\t', header=True, row_names=1)
    return rsnippets.numeric2real(data)


def transpose_file_contents(fileName, rowLabel):
    '''
    read in a file to a dataframe, transpose, and output
    use this instead of R transforms b/c R will concatenate
    an "X" to gene ids starting with a number
    '''
    with open(fileName, 'r') as f:
        content = [line.rstrip().split() for line in f]

    header = True
    with open(fileName, 'w') as f:
        for line in zip(*content):
            if header:
                line = list(line)
                line[0] = rowLabel
                line = tuple(line)
                header = False

            # b/c rpy2 replaces '-' in gene symbols with '.'
            if '.' in line[0]:
                line = list(line)
                line[0] = line[0].replace('.', '-')
                line = tuple(line)

            # b/c R tacks an X on to gene names that start
            # with a #
            if re.search('^X\d', line[0]) is not None:
                print(line[0], file=stderr)
                line = list(line)
                line[0] = re.sub('^X', '', line[0])
                line = tuple(line)

            print('\t'.join(line), file=f)



# pylint: disable=invalid-name
'''
manage colors
'''

import logging
from random import randint
from matplotlib import colors as ref_colors


class Colors(object):
    '''
    manage colors for modules
    '''
    def __init__(self):
        # base colors taken from WGCNA standard coloring for modules
        self.logger = logging.getLogger('iterativeWGCNA.Colors')

        self.base_colors = ['turquoise', 'blue', 'brown', 'yellow', 'green',
                            'red', 'black', 'pink', 'magenta', 'purple',
                            'greenyellow', 'tan', 'salmon', 'cyan', 'midnightblue',
                            'lightcyan', 'lightgreen', 'lightyellow', 'royalblue',
                            'darkred', 'darkgreen', 'darkturquoise', 'orange',
                            'darkorange', 'skyblue', 'saddlebrown', 'steelblue',
                            'paleturquoise', 'violet', 'darkolivegreen',
                            'darkmagenta']

        self.used_colors = []


    def assign_color(self, n):
        '''
        assigns a color
        if n <= len(base_colors) assigns the base color whose index is n - 1
        else generates a random color
        '''
        color = None
        if n <= len(self.base_colors):
            color = ref_colors.cnames[self.base_colors[n - 1]] # get hex representation
        else:
            color = self.__generate_random_color()

        self.used_colors.append(color)
    
        return color


    def __generate_random_color(self):
        '''
        generate a random color
        '''
        color = '#' + '%06X' % randint(0, 0xFFFFFF)
        while color in self.used_colors:
            color = '#' + '%06X' % randint(0, 0xFFFFFF)
        return color



# pylint: disable=invalid-name
# pylint: disable=no-self-use
# pylint: disable=too-many-instance-attributes
# pylint: disable=redefined-variable-type

'''
manage network (for summaries)
'''

from __future__ import print_function
from __future__ import with_statement

import logging

from collections import OrderedDict

import rpy2.robjects as ro

from .colors import Colors
from .wgcna import WgcnaManager
from .r.manager import RManager
from .r.imports import grdevices, base, rsnippets

from .eigengenes import Eigengenes

class Network(object):
    '''
    a network contains the classified genes
    and their assignments
    hash of modules and properties
    for summary purposes
    '''

    def __init__(self, args):
        self.logger = logging.getLogger('iterativeWGCNA.Network')
        self.args = args
        self.eigengenes = None

        self.genes = None
        self.modules = None
        self.classifiedGenes = None
        self.profiles = None
        self.kME = None
        self.membership = None

        # self.graph = None
        self.geneColors = None
        self.adjacency = None
        self.weightedAdjacency = None # weighted by shared membership


    def build(self, genes, eigengenes):
        '''
        build from iterativeWGCNA result in memory
        '''
        self.eigengenes = eigengenes
        self.genes = genes.get_genes()
        self.classifiedGenes = genes.get_classified_genes()
        self.profiles = genes.profiles
        self.kME = genes.get_gene_kME() # TODO -- fix this -- this function has changed
        self.membership = genes.get_gene_membership()

        self.modules = genes.get_modules()
        self.__initialize_module_properties()

        self.__assign_colors()
        self.__generate_weighted_adjacency()


    def __assign_colors(self):
        self.__generate_module_colors()
        self.geneColors = OrderedDict((gene, None) for gene in self.genes)
        self.assign_gene_colors()


    def __initialize_module_properties(self):
        '''
        transform module list into dict with placeholders
        for color, kIn, and kOut
        and values for size
        '''
        self.modules = OrderedDict((module,
                                    {'color':None,
                                     'kIn':0,
                                     'kOut':0,
                                     'size':self.__get_module_size(module),
                                     'density':0}) \
                                       for module in self.modules)
        self.modules.update({'UNCLASSIFIED': {'color':None, 'kIn': 0,
                                              'kOut': 0,
                                              'size': self.__get_module_size('UNCLASSIFIED'),
                                              'density': 0.0}})


    def build_from_file(self, profiles, adjacency=True):
        '''
        initialize Network from iterativeWGCNA output found in path
        '''
        self.profiles = profiles
        self.genes = self.profiles.genes()

        # when membership is loaded from file, modules and classified
        # genes are determined as well
        self.__load_membership_from_file(self.args.preMerge)
        self.__load_kme_from_file(self.args.preMerge)
        warning("done")
        
        self.eigengenes = Eigengenes()
        self.eigengenes.load_matrix_from_file("eigengenes-final.txt")
        if adjacency:
            self.__initialize_module_properties()
            self.__assign_colors()
            self.__generate_weighted_adjacency()


    def __load_membership_from_file(self, preMerge):
        '''
        loads membership assignments from file
        and assembles list of classified genes
        and determines list of unique modules
        '''
        fileName = "membership.txt"
        membership = ro.DataFrame.from_csvfile(fileName, sep='\t',
                                               header=True, row_names=1, as_is=True)

        finalIndex = membership.names.index('final')
        if preMerge:
            finalIndex = finalIndex - 1 

        self.membership = OrderedDict((gene, None) for gene in self.genes)
        self.classifiedGenes = []
        self.modules = []

        for g in self.genes:
            module = membership.rx(g, finalIndex)[0]
            self.modules.append(module)
            self.membership[g] = module
            if module != 'UNCLASSIFIED':
                if 'p' not in module:
                    self.logger.debug("Module: " + module + "; Gene: " + g)
                self.classifiedGenes.append(g)

        self.modules = list(set(self.modules)) # gets unique list of modules


    def __load_kme_from_file(self, preMerge):
        '''
        loads kME to assigned module from file
        '''
        fileName = "eigengene-connectivity.txt"
        kME = ro.DataFrame.from_csvfile(fileName, sep='\t',
                                        header=True, row_names=1, as_is=True)

        finalIndex = kME.names.index('final')
        if preMerge:
            finalIndex = finalIndex - 1
            
        self.kME = OrderedDict((gene, None) for gene in self.genes)
        for g in self.genes:
            self.kME[g] = kME.rx(g, finalIndex)[0]


    def summarize_network(self):
        '''
        generate summary figs and data
        '''
        self.__plot_eigengene_overview()
        self.__plot_summary_views()
        self.__summarize_network_modularity()
        self.__write_module_summary()


    def summarize_module(self, module):
        '''
        generate summare info for the specified module
        '''
        self.__plot_module_overview(module)


    def __plot_module_overview(self, module):
        '''
        plot heatmap, dendrogram, and eigengene
        '''
        grdevices().pdf(module + "-summary.pdf")
        self.plot_module_eigengene(module)
        self.plot_module_kME(module)
        self.plot_module_heatmap(module)
        grdevices().dev_off()


    def plot_module_heatmap(self, module):
        '''
        plot module heatmap
        '''

        members = self.__get_module_members(module)
        expression = self.profiles.gene_expression(members)
        manager = RManager(expression, None)
        manager.heatmap()


    def plot_module_eigengene(self, module):
        '''
        barchart illustrating module eigengene
        '''
        eigengene = self.eigengenes.get_module_eigengene(module)

        params = {}
        params['height'] = base().as_numeric(eigengene)

        limit = max(abs(base().max(eigengene)[0]), abs(base().min(eigengene)[0]))
        ylim = [-1 * limit, limit]
        params['ylim'] = ro.IntVector(ylim)

        colors = ["red" if e[0] > 0 else "blue" for e in eigengene]
        params['col'] = ro.StrVector(colors)

        params['border'] = ro.NA_Logical
        params['las'] = 2
        params['names.arg'] = ro.StrVector(self.eigengenes.samples())
        params['cex.names'] = 0.6
        params['main'] = "Eigengene: " + module
        manager = RManager(eigengene, params)
        manager.barchart()


    def plot_module_kME(self, module):
        '''
        plots module eigengene connectivity (kME)
        '''
        members = self.__get_module_members(module)
        kME = [kME for gene, kME in self.kME.items() if gene in members]

        manager = RManager(kME, None)
        manager.histogram(self.args.wgcnaParameters['minKMEtoStay'],
                          {'main':"Member kME: " + module,
                           'xlab': "Eigengene Connectivity (kME)",
                           'ylab': "N Genes",
                           'breaks': base().seq(0, 1, 0.1)})



    def __generate_module_colors(self):
        '''
        generate module color map
        using standard colors for first
        few modules then random for anything more than first 25
        '''

        colors = Colors()
        n = 0 # counter for module num
        for m in self.modules:
            n = n + 1
            color = colors.assign_color(n)
            self.modules[m]['color'] = color

        self.modules["UNCLASSIFIED"]['color'] = '#D3D3D3'


    def assign_gene_colors(self):
        '''
        assign colors to genes according to module membership
        '''
        for g in self.genes:
            self.geneColors[g] = self.modules[self.membership[g]]['color']


    def get_gene_colors(self, targetGenes):
        '''
        retrieve colors for specified gene list
        '''
        colors = OrderedDict((gene, color) for gene, color in self.geneColors.items() \
                                 if gene in targetGenes)
        return colors


    def get_gene_membership(self, targetGenes):
        '''
        retrieve membership for specified gene list
        '''
        colors = OrderedDict((gene, membership) for gene, membership in self.membership.items() \
                                 if gene in targetGenes)
        return colors


    def __plot_eigengene_overview(self):
        '''
        plots eigengene graphs to single pdf
        '''
        grdevices().pdf("eigengene-overview.pdf")
        self.plot_eigengene_network()
        self.plot_eigengene_heatmap()
        grdevices().dev_off()


    def plot_eigengene_network(self):
        '''
        wrapper for plotting the eigengene network to pdf
        '''
        manager = WgcnaManager(self.eigengenes.matrix, None)
        manager.plot_eigengene_network()


    def plot_eigengene_heatmap(self):
        '''
        plot a heatmap of the eigengenes
        '''
        manager = RManager(self.eigengenes.matrix, None)
        # manager.heatmap(params={'scale':'none'})
        manager.heatmap()

        
    def plot_network_summary(self, genes, title, filename):
        '''
        wrapper for WGCNA summary network view (heatmap + dendrogram)
        '''

        expression = self.profiles.gene_expression(genes)
        membership = self.get_gene_membership(genes)
        # colors = self.get_gene_colors(genes)
        
        manager = WgcnaManager(expression, self.args.wgcnaParameters)
        manager.set_module_colors(self.modules)
        
        grdevices().pdf(filename)
        manager.plot_network_heatmap(membership, title) # plot_network_overview(colors, title)
        grdevices().dev_off()


    def __plot_summary_views(self):
        '''
        plot summary heatmaps/views
        '''
        if self.args.generateNetworkSummary == 'input' \
           or self.args.generateNetworkSummary == 'all':
            self.plot_network_summary(self.genes,
                                      "All Genes (incl. unclassified)",
                                      "input-block-diagram.pdf")

        if self.args.generateNetworkSummary == 'network' \
           or self.args.generateNetworkSummary == 'all':
            self.plot_network_summary(self.classifiedGenes,
                                      "Network (classified genes)",
                                      "network-block-diagram.pdf")


    def __get_module_members(self, targetModule):
        '''
        get genes in targetModule
        '''
        return [gene for gene, module in self.membership.items() if module == targetModule]


    def __generate_weighted_adjacency(self):
        '''
        gene x gene weight matrix with matrix[r][c] = 1
        if genes r & c are in the same module; for
        weighted graph viz and to simply
        calc of in/out degree
        '''

        manager = WgcnaManager(self.profiles.gene_expression(self.classifiedGenes),
                               self.args.wgcnaParameters)
        manager.adjacency('signed', True, True) # signed, but filter negatives & self-refs
        self.adjacency = base().as_data_frame(manager.adjacencyMatrix)
        self.weightedAdjacency = self.adjacency

        for m in self.modules:
            if m == 'UNCLASSIFIED':
                continue

            members = self.__get_module_members(m)
            for r in members:
                indexR = self.weightedAdjacency.names.index(r)
                for c in members:
                    if r == c:
                        continue
                    indexC = self.weightedAdjacency.names.index(c)
                    adj = self.adjacency[indexR][indexC]
                    if adj > 0:
                        self.weightedAdjacency[indexR][indexC] = adj + 1


    def calculate_degree_modularity(self, targetModule):
        '''
        calculates in degree (kIn) and out degree (kOut)
        for the target module
        '''
        members = self.__get_module_members(targetModule)

        degree = rsnippets.degree(self.adjacency, ro.StrVector(members),
                                  self.args.edgeWeight)
        self.modules[targetModule]['kIn'] = int(degree.rx2('kIn')[0])
        self.modules[targetModule]['kOut'] = int(degree.rx2('kOut')[0])
        size = self.modules[targetModule]['size']
        self.modules[targetModule]['density'] = float(self.modules[targetModule]['kIn'])/(float(size) * (float(size) - 1.0)/2.0)


    def __summarize_network_modularity(self):
        '''
        summarize modularity of network
        calculates in degree (kIn) and out degree (kOut) per module
        '''
        for m in self.modules:
            if m != 'UNCLASSIFIED':
                self.calculate_degree_modularity(m)


    def __get_module_size(self, targetModule):
        '''
        return # of module members for target module
        '''
        return len(self.__get_module_members(targetModule))


    def __write_module_summary(self):
        '''
        writes network modularity to a file
        '''
        fileName = 'module-summary.txt'
        with open(fileName, 'w') as f:
            header = ('module', 'size', 'color',
                      'kOut', 'avg_node_kOut',
                      'kIn',
                      'module_density', 'kIn2kOut_ratio')
            print('\t'.join(header), file=f)
            for m in self.modules:
                kIn = self.modules[m]['kIn']
                kOut = self.modules[m]['kOut']
                size = self.modules[m]['size']
                avg_kOut = "{0:0.0f}".format(round(float(kOut) / float(size)))
                ratio = "{0:0.2f}".format(float(kIn) / float(kOut)) \
                         if kOut != 0 else 'NA'

                print(m,
                      size,
                      self.modules[m]['color'],
                      kOut, avg_kOut,
                      kIn,
                      "{0:0.1f}".format(self.modules[m]['density']),
                      ratio,
                      sep='\t', file=f)


    def export_cytoscape_json(self):
        '''
        creates and saves a cytoscape json file
        '''
        filterLevel = self.args.edgeWeight

        if self.weightedAdjacency is None:
            self.__generate_weighted_adjacency()




'''
imports from R; wrapped in functions
to ensure warning messages go to the R log
'''
from rpy2.robjects.packages import importr, SignatureTranslatedAnonymousPackage
from .snippets import FUNCTIONS

rsnippets = SignatureTranslatedAnonymousPackage(FUNCTIONS, 'rsnippets')

def base():
    return importr('base')


def wgcna():
    return importr('WGCNA')


def stats():
    return importr('stats')


def graphics():
    return importr('graphics')


def grdevices():
    return importr('grDevices')


def pheatmap():
    return importr('pheatmap')




'''Run iterativeWGCNA'''

# Installation workaround - see README
# import readline

from iterativeWGCNA.cmlargs import parse_command_line_args
from iterativeWGCNA.iterativeWGCNA import IterativeWGCNA

if __name__ == '__main__':
    cmlArgs = parse_command_line_args()
    alg = IterativeWGCNA(cmlArgs)
    alg.run()

__author__ = 'Emily Greenfest-Allen'
__copyright__ = 'Copyright 2016, University of Pennsylvania'



# pylint: disable=invalid-name
# pylint: disable=unused-import
'''
manage genes
'''
from __future__ import print_function

import logging
from collections import OrderedDict
from collections import Counter

import rpy2.robjects as ro

# from .expression import Expression
from .analysis import calculate_kME
from .eigengenes import Eigengenes
from .r.imports import wgcna, stats, base, rsnippets, grdevices
from .io.utils import xstr
from .r.manager import RManager

class Genes(object):
    '''
    track input genes and their properties, including
    expression profiles, module membership, eigengene
    connectivity
    '''

    def __init__(self, exprData, debug=False):
        '''
        initialize an OrderedDict of genes
        from the row.names of the expression
        data set
        '''
        self.logger = logging.getLogger('iterativeWGCNA.Genes')
        self.profiles = exprData
        self.genes = OrderedDict((geneId, {'module': 'UNCLASSIFIED',
                                           'kME':float('NaN'),
                                           'iteration': None})
                                 for geneId in self.profiles.genes())

        self.size = len(self.genes)
        self.iteration = None
        self.debug = debug


    def get_module(self, gene):
        '''
        returns the assigned module for a gene
        '''
        return self.genes[gene]['module']


    def __is_classified(self, gene):
        '''
        returns true if the feature is classified
        '''
        return self.get_module(gene) != 'UNCLASSIFIED'


    def __update_module(self, gene, module):
        '''
        update gene module
        do not add new genes
        '''
        if gene in self.genes:
            self.genes[gene]['module'] = module
            return True
        else:
            return False


    def __update_classified_iteration(self, gene, iteration):
        '''
        set the iteration during which
        a gene was first classified
        '''
        if gene in self.genes:
            self.genes[gene]['iteration'] = iteration
            return True
        else:
            return False


    def update_membership(self, genes, blocks):
        '''
        fetches new module membership from WGCNA
        blocks and updates relevant genes
        '''
        modules = rsnippets.extractModules(blocks, ro.StrVector(genes))
        # if the feature is in the subset
        # update, otherwise leave as is
        for gene in genes:
            # .rx returns a FloatVector which introduces
            # a .0 to the numeric labels when converted to string
            # which needs to be removed
            # note: R array starts at index 1, python at 0
            module = str(modules.rx(gene, 1)[0]).replace('.0', '')
            if module in ('0', 'grey'):
                module = 'UNCLASSIFIED'
            else:
                module = self.iteration + '_' + 'M' + str(module)
                self.__update_classified_iteration(gene, self.iteration)
            self.__update_module(gene, module)

        return None


    def copy_membership(self, source):
        '''
        updates membership from another Genes object
        '''
        sourceMembership = source.get_gene_membership()
        for gene, module in sourceMembership.items():
            self.__update_module(gene, module)


    def __extract_iteration_genes(self, targetIteration):
        '''
        get genes classified during specified interation
        '''
        assignedIterations = self.__extract_classified_iteration()
        return [gene for gene, iteration in assignedIterations.items()
                if iteration == targetIteration]


    def __extract_classified_iteration(self):
        '''
        get classified iteration as an ordered dict
        '''
        return OrderedDict((gene, membership['iteration'])
                           for gene, membership in self.genes.items())


    def __extract_modules(self):
        '''
        extract module membership as an ordered dict
        '''
        return OrderedDict((gene, membership['module']) for gene, membership in self.genes.items())


    def get_gene_membership(self, genes=None):
        '''
        public facing method for getting gene membership; returns a
        gene -> membership hash

        if gene list is provided, return only the membership assignment
        for the provided genes
        '''
        if genes is None:
            return self.__extract_modules()
        else:
            return OrderedDict((gene, module) for gene, module in self.__extract_modules().items() if gene in genes)


    def get_gene_kME(self):
        '''
        public facing method for getting all gene kMEs
        '''
        return self.__extract_kME()


    def get_iteration_kME(self, iteration):
        '''
        return kME for all assignments made
        during current iteration
        '''
        iterationGenes = self.__extract_iteration_genes(iteration)
        geneKME = self.__extract_kME()
        return [kME for gene, kME in geneKME.items() if gene in iterationGenes]


    def get_module_kME(self, targetModule):
        '''
        get all kME values in a module
        '''
        membership = self.get_module_members(targetModule)
        memberKME = self.__extract_kME()
        return [kME for gene, kME in memberKME.items() if gene in membership]


    def get_kME(self, gene):
        '''
        returns the assigned kME for a gene
        '''
        return self.genes[gene]['kME']


    def __extract_kME(self):
        '''
        extract eigengene connectivity (kME)
        as an ordered dict
        '''
        return OrderedDict((gene, membership['kME']) for gene, membership in self.genes.items())


    def __update_kME(self, gene, kME):
        '''
        update gene eigengene connectivity (kME)
        do not add new genes
        '''
        if gene in self.genes:
            self.genes[gene]['kME'] = kME
            return True
        else:
            return False


    def __update_module_kME(self, module, eigengene, genes=None):
        '''
        update member gene eigengene connectivity (kME)
        for specified module and eigengene
        '''
        members = self.get_module_members(module)
        memberKME = calculate_kME(self.profiles.gene_expression(members),
                                  eigengene, False)

        for gene in memberKME.rownames:
            if genes is not None:
                if gene in genes:
                    self.__update_kME(gene, round(memberKME.rx(gene, 1)[0], 2))


    def update_kME(self, eigengenes, genes=None):
        '''
        update module kME given its eigengene
        '''
        modules = self.get_modules()
        for m in modules:
            moduleEigengene = eigengenes.get_module_eigengene(m)
            self.__update_module_kME(m, moduleEigengene, genes)


    def write(self, prefix='', iteration=None):
        '''
        writes the membership and eigengene connectivity
        to files
        filtering for specific iteration if specified
        '''
        summaryGenes = None
        if iteration is None:
            summaryGenes = self.genes
        else:
            iterationGenes = self.__extract_iteration_genes(iteration)
            summaryGenes = OrderedDict((gene, membership) for gene, membership
                                       in self.genes.items()
                                       if gene in iterationGenes
                                       and membership['module'] != 'UNCLASSIFIED')

        with open(prefix + 'membership.txt', 'w') as f:
            print('\t'.join(('Gene', 'Module', 'kME')), file=f)
            for g in summaryGenes:
                print('\t'.join((g, self.genes[g]['module'], xstr(self.genes[g]['kME']))), file=f)
        return None


    def write_iteration_counts(self, prefix=''):
        '''
        print iteration summary
        '''
        with open(prefix + 'summary.txt', 'w') as f:
            print('\t'.join(('N Input Genes', 'N Classified Genes',
                             'N Residual Genes', 'N Detected Modules')), file=f)
            numClassifiedGenes = self.count_classified_genes()
            print('\t'.join((str(self.size),
                             str(numClassifiedGenes),
                             str(self.size - numClassifiedGenes),
                             str(self.count_modules(self.get_classified_genes())))), file=f)


    def plot_kme_histogram(self, iteration, prefix='', vline=0.80):
        '''
        generate kme histogram for genes classified in
        current iteration
        '''
        kmeVector = None
        if 'final' in prefix or 'merge' in prefix:
            classifiedGenes = self.get_classified_genes()
            geneKME = self.__extract_kME()
            kmeVector = [kME for gene, kME in geneKME.items() if gene in classifiedGenes]
        else:
            kmeVector = self.get_iteration_kME(iteration)

        if kmeVector is not None:
            if len(kmeVector) != 0:
                manager = RManager(kmeVector)
                grdevices().pdf(prefix + "kme_histogram.pdf")
                manager.histogram(vline, {'main': 'Gene -> Assigned Module kME for iteration ' + iteration,
                                          'xlab': 'kME', 'ylab':'Gene Count'})
                grdevices().dev_off()


    def count_module_members(self, genes=None):
        '''
        counts the number of genes per module
        and returns a dict of module -> gene count
        if a list of genes is provided, only counts within
        the specified gene list
        '''
        membership = self.__extract_modules()
        if genes is not None:
            membership = {gene:module for gene, module in membership.items() if gene in genes}
        return Counter(membership.values())


    def count_classified_genes(self, genes=None):
        '''
        counts and return the number of classified genes
        if a list of genes is provided, only counts within
        the specified gene list
        '''
        membership = self.__extract_modules()
        if genes is not None:
            membership = {gene:module for gene, module in membership.items() if gene in genes}
        classified = [gene for gene, module in membership.items()
                      if module != 'UNCLASSIFIED']

        return len(classified)



    def get_classified_genes(self, genes=None):
        '''
        gets the list of classified genes
        if a list of genes is provided, only returns
        genes within the specified list
        '''
        membership = self.__extract_modules()
        if genes is not None:
            membership = OrderedDict((gene, module) for gene, module in membership.items()
                                     if gene in genes)
        classifiedGenes = [gene for gene, module in membership.items()
                           if module != 'UNCLASSIFIED']

        return classifiedGenes



    def get_unclassified_genes(self):
        '''
        get unclassified genes
        '''
        membership = self.__extract_modules()
        unclassifiedGenes = [gene for gene, module in membership.items()
                             if module == 'UNCLASSIFIED']
        return unclassifiedGenes



    def count_modules(self, genes=None):
        '''
        counts the number of modules (excluding unclassified)
        if a list of genes is provided, only counts within
        the specified gene list
        '''
        moduleCount = self.count_module_members(genes)
        return len(moduleCount) - 1 if 'UNCLASSIFIED' in moduleCount else len(moduleCount)


    def remove_small_modules(self, minModuleSize):
        '''
        checks membership counts and removes
        any modules that are too small
        by updating gene membership to UNCLASSIFIED and
        setting eigengene connectivity (kME) to NaN
        '''
        memberCount = self.count_module_members()

        for g in self.genes:
            geneModule = self.get_module(g)
            if memberCount[geneModule] < minModuleSize:
                self.__update_module(g, 'UNCLASSIFIED')
                self.__update_kME(g, float('NaN'))
                self.__update_classified_iteration(g, None)


    def get_modules(self, genes=None):
        '''
        gets list of unique modules from gene membership assignments
        '''
        # get unique members by converting values to a set
        membership = set(self.__extract_modules().values())
        membership.discard('UNCLASSIFIED')
        return list(membership)

    
    def get_module_members(self, targetModule):
        '''
        get list of module member genes
        '''
        membership = self.__extract_modules()
        return [gene for gene, module in membership.items() if module == targetModule]


    def get_genes(self):
        '''
        return list of all genes
        '''
        return [gene for gene in self.genes]


    def evaluate_fit(self, minKMEtoStay, genes=None):
        '''
        evaluate fit of each gene to its assigned
        module, unclassifying if the fit is below the
        minimum KME to stay
        if a gene list is provided, only evaluates the
        specified genes
        '''

        if genes is None:
            genes = self.profiles.genes()

        for g in genes:
            module = self.get_module(g)
            kME = self.get_kME(g)

            if module == 'UNCLASSIFIED':
                self.__update_kME(g, float('NaN'))
                self.__update_classified_iteration(g, None)

            if kME < minKMEtoStay:
                self.__update_module(g, 'UNCLASSIFIED')
                self.__update_kME(g, float('NaN'))
                self.__update_classified_iteration(g, None)


    def merge_close_modules(self, eigengenes, cutHeight):
        '''
        merge close modules based on similarity between
        eigengenes

        return updated eigengene object
        '''

        # repeat until no more merges are possible
        noMergesFound = False
        mergeCount = 0
        modules = self.get_modules()
        classifiedGenes = self.get_classified_genes()
        classifiedGeneProfiles = self.profiles.gene_expression(classifiedGenes)
        while not noMergesFound:
            # compare modules, finding min dissimilarity
            similarity = eigengenes.similarity(None)
            closeModules = rsnippets.findCloseModules(similarity, cutHeight)
            if closeModules != ro.NULL:
                m1 = closeModules.rx2('m1')[0]
                m2 = closeModules.rx2('m2')[0]
                dissimilarity = closeModules.rx2('dissimilarity')[0]
                mergeCount = mergeCount + 1
                self.logger.info("Merging " + m1 + " into " + m2
                                 + " (D = " + str(dissimilarity) + ")")

                memberGenes = self.get_module_members(m1)
                for g in memberGenes:
                    self.__update_module(g, m2)
                    self.__update_classified_iteration(g, 'FINAL_MERGE')
                self.__update_module_kME(m1, eigengenes.get_module_eigengene(m2))

                modules = self.get_modules()
                classifiedGeneMembership = self.get_gene_membership(classifiedGenes)
                if self.debug:
                    self.logger.debug("Getting module assignments for classified genes")
                    self.logger.debug(classifiedGeneMembership)

                eigengenes.recalculate(classifiedGeneProfiles,
                                       classifiedGeneMembership)

            else:
                noMergesFound = True

        self.logger.info("Done merging close modules: " + str(mergeCount) + " modules merged.")
        self.logger.info("Retained " + str(len(modules)) + " modules after merge.")

        return eigengenes


    def reassign_to_best_fit(self, eigengenes, reassignThreshold, minKMEtoStay):
        '''
        Evaluate eigengene connectivity (kME)
        for each feature against the eigengenes for each
        of the final modules found by iWGCNA.
        If kME(module) > kME(assigned_module)
        and the p-value <= the reassignThreshold (of WGCNA
        parameters) then reassign the module
        membership of the feature.

        returns a count of the number of reassigned genes
        '''
        count = 0
        modules = self.get_modules()
        for m in modules:
            # calculate kME of all genes to the module eigengene
            moduleEigengene = eigengenes.get_module_eigengene(m)
            moduleKME = calculate_kME(self.profiles.expression(), moduleEigengene, True)

            # for each gene not assigned to the current module, test fit
            for g in self.genes:
                currentModule = self.get_module(g)
                if currentModule != m:
                    kME = self.get_kME(g)
                    newKME = round(moduleKME.rx2('cor').rx(g, 1)[0], 2)
                    pvalue = moduleKME.rx2('p').rx(g, 1)[0]

                    if (currentModule == "UNCLASSIFIED" \
                        and newKME >= minKMEtoStay) \
                        or (newKME > kME \
                        and pvalue < reassignThreshold):

                        self.__update_module(g, m)
                        self.__update_kME(g, newKME)
                        count = count + 1

        return count


    def load_membership(self, fileName=None):
        '''
        loads membership
        '''
        if fileName is None:
            fileName = "final-membership.txt"

        membership = ro.DataFrame.from_csvfile(fileName, sep='\t',
                                               header=True, row_names=1, as_is=True)

        if self.debug:
            self.logger.debug("Loaded membership from file " + fileName + "; see R-log")
            ro.r("print('Loaded membership from file -- head of file:')")
            self.logger.debug(membership.head())

        index = membership.names.index('Module') + 1 # add 1 b/c of python/rpy2/R inconsistency

        if self.debug:
            self.logger.debug("Adjusted index of Module column: " + str(index))

        classifiedCount = 0
        unclassifiedCount = 0
        for g in self.genes:
            gStr = ro.StrVector([str(g)])
            # strange, but necessary so that rpy2 will treat numeric gene ids as strings
            # python str() conversion did not work

            module = membership.rx(gStr[0], index)[0]

            if module == 'UNCLASSIFIED':
                unclassifiedCount = unclassifiedCount + 1
            else:
                classifiedCount = classifiedCount + 1
            self.__update_module(g, module)

        self.logger.info("Loaded " + str(classifiedCount) + " classified genes")
        self.logger.info("Loaded " + str(unclassifiedCount) + " unclassified genes")



'''
functions in support of data analysis
'''

# TODO move to RManager or wgcnaManager

from .r.imports import wgcna, stats, base

def calculate_kME(expr, eigengene, calculateP):
    '''
    calculates eigengene connectivity
    between an eigengene and expression data set
    '''
    if calculateP:
        correlation = wgcna().corAndPvalue(base().t(expr), base().t(eigengene))
    else:
        correlation = base().as_data_frame(stats().cor(base().t(expr), base().t(eigengene)))
    return correlation



#!/usr/bin/env python2.7
"""
uses rpy2 python library to create a namespace for R functions underlying iterativeWGCNA
"""

FUNCTIONS = """

# convert numeric data frame to real
numeric2real <- function(df) {
     df * 1.0
}

# return 1-matrix
dissMatrix <- function(df) {
    1.0 - df
}

# add a constant value
add <- function(df, value) {
    df + value
}

# return power-weighted matrix
powerWeightMatrix <- function(df, power) {
    df^power
}


# set values < thresshold to 0
filterByThreshold <- function(df, threshold) {
    df[df < threshold] <- 0
    df
}


# set value of matrix diagonal
diag <- function(df, value) {
    diag(df) <- value
    df
}

# wrapper for save object b/c doesn't seem to work with rpy2
saveObject <- function(obj, objName, file) {
   assign(objName, obj)
   save(list=c(objName), file = file)
}

saveBlockResult <- function(blocks, profiles, file) {
   assign('blocks', blocks)
   assign('expression', profiles)
   save(list=c('blocks', 'expression'), file = file)
}

# calculate degree summary for module genes
degree <- function(adjMatrix, members, threshold) {
    adjSubset <- adjMatrix[members, members]
    inDegree = sum(adjSubset >= threshold) / 2
    adjSubset <- adjMatrix[members,  !names(adjMatrix) %in% members]
    outDegree <- sum(adjSubset >= threshold)
    list(kIn=inDegree, kOut=outDegree)
}


# find two closest modules given a similarity threshold
findCloseModules <- function(similarityMatrix, cutHeight) {
     returnVal <- NULL
     d <- 1 - similarityMatrix
     comparison <- d[d > 0 & d <= cutHeight]
     modulesFound <- sum(comparison) > 0
print(cutHeight)

     if (modulesFound) {
         # indexes of closest modules
         indexes <- which(d == min(comparison), arr.ind = TRUE)
         returnVal <- list(m1 = row.names(d)[indexes[1,1]], m2 = row.names(d)[indexes[1,2]], dissimilarity = d[indexes[1,1], indexes[1,2]])
     }
    returnVal
}


# given WGCNA blocks, extracts and transposes eigengene matrix
# labels columns (samples)
# cleans up module names (removes the "ME")

extractEigengenes <- function(iteration, blocks, sampleNames) {
    eigengenes <- as.data.frame(t(blocks$MEs))
    colnames(eigengenes) <- sampleNames
    eigengenes <- eigengenes[row.names(eigengenes) != "ME0" & row.names(eigengenes) != "MEgrey", ]
    row.names(eigengenes) <- gsub("ME", paste(iteration, "_M", sep=""), row.names(eigengenes))
    eigengenes
}


# extract eigengens from list object output
# from moduleEigengenes function
# label columns (samples)
# clean up module names (remove the "ME")
extractRecalculatedEigengenes <- function(elist, sampleNames) {
   eigengenes <- as.data.frame(t(elist$eigengenes))
   colnames(eigengenes) <- sampleNames
    row.names(eigengenes) <- gsub("ME", "", row.names(eigengenes))
    eigengenes
}

# given WGCNA blocks and gene names, returns
# a data frame with modules mapped to gene names
extractModules <- function(blocks, geneNames) {
    as.data.frame(blocks$colors, row.names = geneNames)
}

# extract module members
# does not assume same ordering
extractMembers <- function(module, expr, membership) {
    membership <- unlist(membership)
    membership <- membership[row.names(expr)]
    members <- membership == module
    expr[members, ]
}

# remove unclassified from expression set
removeUnclassified <- function(expr, membership) {
    membership <- unlist(membership)
    membership <- membership[row.names(expr)]
    classified = membership != "UNCLASSIFIED"
    expr[classified, ]
}

# blue, white, red color scale
BlWhRed <- function() {
    colorRampPalette(c("blue", "white", "red"))(100)
}

# white, yellow, red color scale
WhYlRed <- function() {
   colorRampPalette(c("white", "yellow", "red"))(100)
}

# create color scale by passing a string vector of colors
colorScale <- function(colors) {
   colorRampPalette(colors)(100)
}

"""

__author__ = "Emily Greenfest-Allen"
__copyright__ = "Copyright 2016, University of Pennsylvania"



# pylint: disable=invalid-name
# pylint: disable=bare-except
# pylint: disable=broad-except
# pylint: disable=too-many-instance-attributes
'''
main application
'''

from __future__ import print_function

import logging
import sys
import os
from time import strftime

import rpy2.robjects as ro
from .genes import Genes
from .expression import Expression
from .eigengenes import Eigengenes
from .network import Network
from .wgcna import WgcnaManager
from .io.utils import create_dir, read_data, warning, write_data_frame, bulk_gzip
from .r.imports import base, wgcna, rsnippets


class IterativeWGCNA(object):
    '''
    main application

    flag report = True when generating
    result from existing output
    '''

    def __init__(self, args, report=False):
        self.args = args
        create_dir(self.args.workingDir)
        if not report:
            self.__verify_clean_working_dir()

        if report == 'merge':
            self.args.enableWGCNAThreads = False

        self.__initialize_log(report)
        self.logger.info(strftime("%c"))

        self.__initialize_R(report)
        if not report:
            self.__log_parameters()

        if self.args.debug:
            warning("Running in DEBUG mode.")
            warning("Rpy2 will print debugging messages and variable (e.g., matrix/vector) contents to the R log")
            warning("Thus, empty debug statements 'DEBUG:      ' in the iterativeWGCNA log should have a corresponding output in the R log")

        # load expression data and
        # initialize Genes object
        # to store results
        self.profiles = None
        self.__load_expression_profiles()
        self.__log_input_data()
        self.genes = Genes(self.profiles, debug=self.args.debug)
        self.eigengenes = Eigengenes(debug=args.debug)
        self.modules = None # will be hash of module name to color for plotting

        if not report:
            self.passCount = 1
            self.iterationCount = 1
            self.iteration = None # unique label for iteration
            self.algorithmConverged = False
            self.passConverged = False


    def __verify_clean_working_dir(self):
        '''
        verifies that working directory does not contain
        iterativeWGCNA output files
        exits to avoid accidental overwrite of earlier runs
        '''
        conflictingFiles = set(('final-eigengenes.txt', 'final-membership.txt'))
        files = set(os.listdir(self.args.workingDir))
        if len(files.intersection(conflictingFiles)) > 0:
            warning("Working Directory: " + self.args.workingDir \
                               + " contains final output from a prior run of iterativeWGCNA.  Exiting...")
            sys.exit(1)


    def run_pass(self, passGenes):
        '''
        run a single pass of iterative WGCNA
        (prune data until no more residuals are found)
        '''

        passDirectory = 'pass' + str(self.passCount)
        create_dir(passDirectory)
        write_data_frame(self.profiles.gene_expression(passGenes),
                         os.path.join(passDirectory, 'initial-pass-expression-set.txt'),
                         'Gene')

        iterationGenes = passGenes

        while not self.passConverged:
            self.run_iteration(iterationGenes)

            moduleCount = self.genes.count_modules(iterationGenes)
            classifiedGeneCount = self.genes.count_classified_genes(iterationGenes)

            self.write_run_summary(len(iterationGenes), classifiedGeneCount)

            # if there are no residuals
            # (classified gene count = number of genes input)
            # then the pass has converged
            if classifiedGeneCount == len(iterationGenes):
                self.passConverged = True
                self.__summarize_classification(passDirectory + '/')
            else:
                # run again with genes classified in current pass
                iterationGenes = self.genes.get_classified_genes(iterationGenes)
                self.iterationCount = self.iterationCount + 1

            # if no modules were detected,
            # then the algorithm has converged
            # exit the pass
            if moduleCount == 0:
                self.algorithmConverged = True
                self.passConverged = True
                self.__log_alogorithm_converged()


    def run_iterative_wgcna(self):
        '''
        run iterative WGCNA
        '''
        if self.args.verbose:
            warning("Beginning iterations")

        # genes involved in current iteration
        passGenes = self.profiles.genes()

        while not self.algorithmConverged:
            self.run_pass(passGenes)
            classifiedGeneCount = self.genes.count_classified_genes(passGenes)
            self.__log_pass_completion()
            self.__log_gene_counts(len(passGenes), classifiedGeneCount)

            if not self.algorithmConverged:
                # set residuals as new gene list
                passGenes = self.genes.get_unclassified_genes()

                # increment pass counter and reset iteration counter
                self.passCount = self.passCount + 1
                self.iterationCount = 1

                # reset pass convergence flag
                self.passConverged = False

        self.iteration = 'FINAL'
        self.genes.iteration = self.iteration
        self.__log_gene_counts(self.genes.size, self.genes.count_classified_genes())
        self.__summarize_classification('final-')

        # output current eigengenes for all modules, not just ones from last pass
        self.eigengenes.load_matrix_from_file('eigengenes.txt')
        modules = self.genes.get_modules()
        self.eigengenes.update_to_subset(modules)
        self.eigengenes.write('final-')

        self.iteration = 'MERGED'
        self.genes.iteration = self.iteration
        self.merge_close_modules()
        self.reassign_genes_to_best_fit_module()

        self.__log_gene_counts(self.genes.size, self.genes.count_classified_genes())

        self.__summarize_classification('merged-' + str(self.args.finalMergeCutHeight) + '-')
        self.eigengenes.write('merged-' + str(self.args.finalMergeCutHeight) + '-')
        os.remove("eigengenes.txt")


    def merge_close_modules_from_output(self):
        '''
        load data from output and remerge
        '''
        self.genes.load_membership()
        self.merge_close_modules('final-')
        self.reassign_genes_to_best_fit_module()
        self.__log_gene_counts(self.genes.size, self.genes.count_classified_genes())
        self.genes.write('adjusted-merge-' + str(self.args.finalMergeCutHeight) + '-')
        self.eigengenes.write('adjusted-merge-' + str(self.args.finalMergeCutHeight) + '-')
        # self.transpose_output_files()


    def summarize_results(self):
        '''
        generate summary output and graphics
        '''
        network = Network(self.args)
        network.build(self.genes, self.eigengenes)
        network.summarize_network()


    def run(self):
        '''
        main function --> makes calls to run iterativeWGCNA,
        catches errors, and logs time
        '''

        try:
            self.run_iterative_wgcna()
            # self.summarize_results() # can cause memory issues so, removing
            self.logger.info('iterativeWGCNA: SUCCESS')
        except Exception:
            if self.logger is not None:
                self.logger.exception('iterativeWGCNA: FAIL')
            else:
                raise
        finally:
            if self.logger is not None:
                self.logger.info(strftime("%c"))


    def reassign_genes_to_best_fit_module(self):
        '''
        use kME goodness of fit to reassign module
        membership
        '''
        if self.args.verbose:
            warning("Making final goodness of fit assessment")

        count = self.genes.reassign_to_best_fit(self.eigengenes,
                                                self.args.wgcnaParameters['reassignThreshold'],
                                                self.args.wgcnaParameters['minKMEtoStay'])
        self.logger.info("Reassigned " + str(count) + " genes in final kME review.")
        if self.args.verbose:
            warning("Reassigned " + str(count) + " genes in final kME review.")


    def merge_close_modules(self, prefix=''):
        '''
        merge close modules based on similiarity in eigengenes
        update membership, kME, and eigengenes accordingly
        '''
        if self.args.verbose:
            warning("Extracting final eigengenes and merging close modules")

        modules = self.genes.get_modules()
        self.__log_final_modules(modules)

        self.eigengenes.load_matrix_from_file(prefix + 'eigengenes.txt')
        self.eigengenes.update_to_subset(modules)

        self.eigengenes = self.genes.merge_close_modules(self.eigengenes,
                                                         self.args.finalMergeCutHeight)


    def run_iteration(self, iterationGenes):
        '''
        run an iteration of blockwise WGCNA
        '''
        self.__generate_iteration_label()

        iterationDir = os.path.join('pass' + str(self.passCount), 'i' + str(self.iterationCount))
        create_dir(iterationDir)

        if self.args.verbose:
            warning("Iteration: " + self.iteration)

        self.genes.iteration = self.iteration
        iterationProfiles = self.profiles.gene_expression(iterationGenes)

        blocks = self.run_blockwise_wgcna(iterationProfiles, iterationDir)
        if not self.args.skipSaveBlocks:
            rsnippets.saveBlockResult(blocks, iterationProfiles,
                                      os.path.join(iterationDir, 'wgcna-blocks.RData'))
            if self.args.gzipTOMs:
                bulk_gzip(iterationDir, 'TOM')


        # update eigengenes from blockwise result
        # if eigengenes are present (modules detected), evaluate
        # fitness and update gene module membership
        self.eigengenes.extract_from_blocks(self.iteration, blocks,
                                            self.profiles.samples())

        if not self.eigengenes.is_empty():
            self.eigengenes.write() # need to keep single file across all iterations
            self.eigengenes.write(iterationDir + '/')

            # extract membership from blocks and calc eigengene connectivity
            self.genes.update_membership(iterationGenes, blocks)
            self.genes.update_kME(self.eigengenes, iterationGenes)
            self.__summarize_classification(os.path.join(iterationDir, 'wgcna-'))

            self.genes.evaluate_fit(self.args.wgcnaParameters['minKMEtoStay'],
                                    iterationGenes)
            self.genes.remove_small_modules(self.args.wgcnaParameters['minModuleSize'])
            self.__summarize_classification(os.path.join(iterationDir, ''), True)


    def __summarize_classification(self, prefix, inclCounts=False):
        '''
        output gene summaries for the iteration
        incl: text summary of iteration, updated gene membership, kme histogram
        '''
        if 'final' in prefix or 'merge' in prefix:
            self.genes.write(prefix)
        else:
            self.genes.write(prefix, self.iteration)
        self.genes.plot_kme_histogram(self.iteration, prefix,
                                      self.args.wgcnaParameters['minKMEtoStay'])
        if inclCounts:
            self.genes.write_iteration_counts(prefix)


    def run_blockwise_wgcna(self, exprData, workingDir):
        '''
        run WGCNA
        '''
        manager = WgcnaManager(exprData, self.args.wgcnaParameters)
        manager.set_parameter('saveTOMFileBase', os.path.join(workingDir, self.iteration + '-TOM'))
        return manager.blockwise_modules()


    def __generate_iteration_label(self):
        '''
        generates the unique label for the iteration
        '''
        self.iteration = 'P' + str(self.passCount) + '_I' + str(self.iterationCount)


    def __load_expression_profiles(self):
        # gives a weird R error that I'm having trouble catching
        # when it fails
        # TODO: identify the exact exception
        try:
            self.profiles = Expression(read_data(self.args.inputFile))
        except:
            self.logger.error("Unable to open input file: " + self.args.inputFile)
            sys.exit(1)


    def __initialize_R(self, logType='run'):
        '''
        initialize R workspace and logs
        '''
        # set working directory
        base().setwd(self.args.workingDir)

        # suppress warnings
        ro.r['options'](warn=-1)

        # r log
        logFile = 'iterativeWGCNA-R.log'
        if logType == 'merge':
            logFile = 'adjust-merge-' + str(self.args.finalMergeCutHeight) + '-' + logFile

        rLogger = base().file(logFile, open='wt')
        base().sink(rLogger, type=base().c('output', 'message'))

        if self.args.enableWGCNAThreads:
            wgcna().enableWGCNAThreads()


    def __initialize_log(self, logType='run'):
        '''
        initialize log by setting path and file format
        '''
        logName = 'iterativeWGCNA.log'
        if logType == 'summary':
            logName = 'summarize-network-' + logName
        elif logType == 'merge':
            logName = 'adjust-merge-' + str(self.args.finalMergeCutHeight) + '-' + logName

        logging.basicConfig(filename=self.args.workingDir + '/' + logName,
                            filemode='w', format='%(levelname)s: %(message)s',
                            level=logging.DEBUG)

        logging.captureWarnings(True)
        self.logger = logging.getLogger(__name__)


    def __log_alogorithm_converged(self):
        '''
        log algorithm convergence
        '''
        message = "No modules detected for iteration " + self.iteration \
                  + ". Classification complete."
        self.logger.info(message)
        if self.args.verbose:
            warning(message)


    def __log_parameters(self):
        '''
        log WGCNA parameter choices and working
        directory name
        '''

        self.logger.info("Working directory: " + self.args.workingDir)
        self.logger.info("Saving blocks for each iteration? "
                         + ("FALSE" if self.args.skipSaveBlocks else "TRUE"))
        self.logger.info("Merging final modules if cutHeight <= " + str(self.args.finalMergeCutHeight))
        self.logger.info("Allowing WGCNA Threads? "
                         + ("TRUE" if self.args.enableWGCNAThreads else "FALSE"))
        self.logger.info("Running WGCNA with the following params:")
        self.logger.info(self.args.wgcnaParameters)

        if self.args.verbose:
            warning("Working directory: " + self.args.workingDir)
            warning("Allowing WGCNA Threads? "
                    + ("TRUE" if self.args.enableWGCNAThreads else "FALSE"))
            warning("Merging final modules if cutHeight <= " + str(self.args.finalMergeCutHeight))
            warning("Running WGCNA with the following params:")
            warning(self.args.wgcnaParameters)


    def __log_input_data(self):
        '''
        log input details
        '''
        self.logger.info("Loaded file: " + self.args.inputFile)
        self.logger.info(str(self.profiles.ncol()) + " Samples")
        self.logger.info(str(self.profiles.nrow()) + " Genes")
        if self.args.verbose:
            warning("Loaded file: " + self.args.inputFile)
            warning(str(self.profiles.ncol()) + " Samples")
            warning(str(self.profiles.nrow()) + " Genes")


    def __log_pass_completion(self):
        '''
        summarize pass in log when convergence condition is met
        '''
        message = "Pass " + str(self.passCount) + " converged on iteration " \
                  + str(self.iterationCount) + "."
        self.logger.info(message)
        if self.args.verbose:
            warning(message)


    def __log_gene_counts(self, initialGeneCount, classifiedGeneCount):
        '''
        log classified gene count
        '''
        message = "FIT: " + str(classifiedGeneCount) + "; RESIDUAL: "
        message = message + str(initialGeneCount - classifiedGeneCount)
        self.logger.info(message)
        if self.args.verbose:
            warning(message)


    def __log_final_modules(self, modules):
        '''
        log modules
        '''
        message = "Found " + str(len(modules)) + " modules."
        self.logger.info(message)
        self.logger.info(modules)
        if self.args.verbose:
            warning(message)
            warning(modules)


    def write_run_summary(self, initial, fit):
        '''
        writes the number of kept and dropped genes at the end of an iteration
        '''
        fileName = 'iterative-wgcna-run-summary.txt'
        try:
            os.stat(fileName)
        except OSError:
            header = ('Iteration', 'Initial', 'Fit', 'Residual')
            with open(fileName, 'a') as f:
                print('\t'.join(header), file=f)
        finally:
            with open(fileName, 'a') as f:
                print('\t'.join((self.iteration, str(initial),
                                 str(fit), str(initial - fit))), file=f)
